#include <fcntl.h>
#include <getopt.h>
#include <net/if.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <errno.h>
#include <linux/netlink.h>
#include <linux/socket.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <vector>
#include <list>
#include <string>
#include <iostream>
#include "ne6x_common.h"

#define HEADER_SIZE             16
#define FACT_PAYLOAD            64 * 44
#define NETLINK_TEST            31
#define MAX_PAYLOAD             1024 * 4 /* maximum payload size */

typedef struct configure {
	char *appName;
	char infName[64];
	uint32_t param[2];
	uint32_t paramCount;
	bool hasInf;
	bool hasParam;
	bool writeData;
	bool readData;
} configure_t;

static configure_t configure = {
	.infName = {0},

	.hasInf = false,
	.writeData = false,
	.readData = false,
};

static void usage(char *appName)
{
	printf("\r\n"
	       "Usage: %s [-h] -i <interface> -f <package> \r\n"
	       " -h             Print this message and exit\r\n"
	       " -i [device]    Interface to use, for example 'eth0'\r\n"
	       " -p [param]  2 param'\r\n"
	       " If none of l/a/n/i specified, then all parts will be updated.\r\n",
	       appName);
}

static int32_t configure_parse_ops(configure_t *cfg, uint32_t argc, char *argv[])
{
	uint32_t c,d;
	char *appName = argv[0];

	if (argc < 2) {
		usage(appName);
		return -1;
	}
	cfg->paramCount = -1;
	cfg->hasInf = false;
	cfg->hasParam = false;
	while (EOF != (c = getopt(argc, argv, "arhi:p:"))) {
		if(c < 0)
			break;
		switch (c) {
		case 'a':
			cfg->writeData = true;
			cfg->paramCount = 2;
			break;
		case 'r':
			cfg->readData = true;
			cfg->paramCount = 1;
			break;		
		}
	}
	if (!(cfg->writeData || cfg->readData)) {
		printf("Please provide write tpl through -a, or provide read tpl through -r\n");
		return -1;
	}
	if ((cfg->writeData && cfg->readData)) {
		printf("Please only provide write tpl through -a, or provide read tpl through -r\n");
		return -1;
	}

	optind = 2;
	while (EOF != (c = getopt(argc, argv, "arhi:p:"))) {
		if(c < 0)
			break;
		switch (c) {
		case 'i':
			sprintf(cfg->infName, "%s", optarg);
			cfg->hasInf = true;
			break;
		case 'p':
			cfg->hasParam = true;
			// std::cout <<"optind" << optind << std::endl;

			for(int i = 0; i < cfg->paramCount; i++)
			{	
				if (strtoul(argv[optind - 1 + i], NULL, 10) >= 0) {
					cfg->param[i] = strtoul(argv[optind - 1 + i], NULL, 10);
				}
				else 
				{
					printf("please check params!\n");
					return -1;
				}
			}

			// for(int i = 0; i < cfg->paramCount; i++)
			// 	std::cout <<"paramCount:" << cfg->param[i] << std::endl;
			
			break;

		case 'h':
			usage(appName);
			cfg->appName = appName;
			return -1;
		// default:
		// 	printf("Unsupported argument!\n");
		// 	return -1;
		}
	}

	if (!(cfg->hasInf && cfg->hasParam)) {
		printf("Please provide interface through -i, provide params through -p\n");
		return -1;
	}

	return 0;
}

static int brust_data(char *p, nlmsghdr *nlh, configure_t *cfg)
{
	uint32_t  sigle_data = 0;
	memcpy(p, (char *)nlh, HEADER_SIZE);               /* fill nlmsg headr */
	memcpy(p + HEADER_SIZE, (char *)cfg->infName, 16); /* fill device name */
	if(cfg->writeData) {
		memcpy(p + HEADER_SIZE + 16, &cfg->param[0], 4);
		memcpy(p + HEADER_SIZE + 20, &cfg->param[1], 4);
	}
	else {
		memcpy(p + HEADER_SIZE + 16, &cfg->param[0], 4);
	}
	return 0;
}

int32_t send_tpl(int argc, char *argv[])
{
	configure_t *cfg = &configure;

	if (0 != configure_parse_ops(cfg, argc, argv)) {
		return -1;
	}
	// return 0;
	
	struct sockaddr_nl src_addr, dest_addr;
	struct nlmsghdr *nlh = NULL;
	int sock_fd, retval;
	struct msghdr msg;
	struct iovec iov;
	int state;
	char send_data[MAX_PAYLOAD];
	int state_smg = 0;
	int dex = 0;
	int count = 1;
	int wait_recvmsg = 1;
	struct netlink_state send_result;

	sock_fd = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_TEST);
	if (sock_fd == -1) {
		printf("error getting socket: %s\n", strerror(errno));
		return -1;
	}
	/* To prepare binding */
	memset(&src_addr, 0, sizeof(src_addr));
	src_addr.nl_family = AF_NETLINK;
	src_addr.nl_pid = getpid();
	src_addr.nl_groups = 0;
	/* Bind */
	retval = bind(sock_fd, (struct sockaddr *)&src_addr, sizeof(src_addr));
	if (retval < 0) {
		printf("bind failed: %s", strerror(errno));
		close(sock_fd);
		return -1;
	}
	/* To orepare create mssage */
	nlh = (struct nlmsghdr *)malloc(NLMSG_SPACE(FACT_PAYLOAD));
	if (!nlh) {
		printf("malloc nlmsghdr error!\n");
		close(sock_fd);
		return -1;
	}
	memset(&dest_addr, 0, sizeof(dest_addr));
	dest_addr.nl_family = AF_NETLINK;
	dest_addr.nl_pid = 0;
	dest_addr.nl_groups = 0;


	nlh->nlmsg_pid = getpid();
	nlh->nlmsg_flags = 1;
	if(cfg->writeData) {
		nlh->nlmsg_len = NLMSG_SPACE(24);
		nlh->nlmsg_type = TEMPL_WRITE;
	}
	else {
		nlh->nlmsg_len = NLMSG_SPACE(20);
		nlh->nlmsg_type = TEMPL_READ;
	}
	/* send message */
	// printf("state_smg\n");
	dex = 0;

	brust_data(send_data, nlh, cfg);

	iov.iov_base = (void *)send_data;
	iov.iov_len = nlh->nlmsg_len;
	/* Create message */
	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (void *)&dest_addr;
	msg.msg_namelen = sizeof(dest_addr);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	state_smg = sendmsg(sock_fd, &msg, 0);
	if (state_smg == -1) {
		printf("get error sendmsg = %s\n", strerror(errno));
	}

	memset(nlh, 0, NLMSG_SPACE(FACT_PAYLOAD));
	// receive message
	printf("waiting received!\n");
	while (wait_recvmsg) {
		state = recvmsg(sock_fd, &msg, 0);
		if (state <= 0 ) {
			printf("waiting recvmsg!\n");
		}
		if (state > 0) {
			send_result = *(struct netlink_state*) NLMSG_DATA((nlmsghdr *)send_data);
			if(send_result.readData[0] && cfg->writeData)
				printf("send message failed!\n");
			else if (cfg->writeData)
				printf("send message sucess!\n");
			
			if(cfg->readData)
                        printf("temp_read:0x%04X value 0x%08X\n",cfg->param[0], send_result.readData[0]);
			wait_recvmsg = 0;
		}
	}
	close(sock_fd);
	free(nlh);
	return 0;
}