/*
 * Copyright (C) 2023 BZWX Co., Ltd. All rights reserved.
 */
#include <fcntl.h>
#include <getopt.h>
#include <net/if.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <errno.h>
#include <linux/netlink.h>
#include <linux/socket.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <vector>
#include <list>
#include <string>
#include <iostream>
#include "ne6x_common.h"

using namespace std;
 
#define MAX_PAYLOAD             1024 * 4 /* maximum payload size */
#define NETLINK_TEST            31
#define PARA_KEY_STRING         " "
#define MAX_SEND                20
#define HEADER_SIZE             16
#define MESSEAGE_LENGTH         28 + 12 + 12
#define FACT_PAYLOAD            64 * 44

typedef struct message {
	char port[16];
	uint32_t src_ip[4];
	uint32_t dest_ip[4];
	uint32_t prot_type;
} message_t;

typedef struct configure {
	char *appName;
	char infName[64];
	char filename[256];

	bool hasInf;
	bool hasFile;
	bool addData;
	bool rmData;
	bool isip6;
} configure_t;

static configure_t configure = {
	.infName = {0},
	.filename = {0},

	.hasInf = false,
	.hasFile = false,
	.addData = false,
	.rmData = false,
};

static int complete_ipv6(char* ipv6, char* complete_v6)
{
    char ipcom[40] = {"0000:0000:0000:0000:0000:0000:0000:0000"};
    char *tmpcomv6[128] = { 0 };
    char *p = NULL, *q = NULL, *q1 = NULL;
    int cnt = 0, i = 0, len = 0, l1 = 0, l2 = 0, l3 = 0, sum = 0, l1tmp = 0;
    // 将IPv6地址用::分开
    // 如果IPv6地址为::
    // 如果不包含::或以::结尾
    // 如果以::开头或::在中间
    p = strchr(ipv6, '/');
    if (p != NULL) {
        *p = '\0';
    }
    p = strstr(ipv6, "::");
    len = strlen(ipcom);
    if (p != NULL) {
        if (p == ipv6) {
            // 地址以::开头
            if (*(p + 2) == '\0') {
                // 地址为::
                memcpy(complete_v6, ipcom, 40);
            } else {
                // 地址为::xxx
                sum = 0;
                q1 = NULL;
                q = p + 2;
                l1 = strlen(q);
                for (i = 0; i < l1; i++) {
                    if (*(p + 1 + l1 -i) != ':') {
                        q1 = strrchr(q, ':');

                        *(ipcom + len - i -1 - sum) = *(p + 1 + l1 -i);
                    } else {
                        l1tmp = strlen(q);
                        sum += (4 - (l1tmp - (q1 - q) -1));
                        *q1 = '\0';
                        cnt += 1;
                    }
                }
            }
        }
        // 地址形式xxx::xxx
        if (p != ipv6) {
            // 处理xxx::
            cnt = 0;
            q = ipv6;
            l3 = p - ipv6;
            for (i = 0; i < l3; i++) {
                if (*(ipv6 + i) != ':') {
                    q1 = strchr(q, ':');
                    *(ipcom + (4 - (q1 - q)) + sum  + i) = *(ipv6 + i);
                } else {
                    sum += (4 - (q1 - q));
                    q1 = q1 + 1;
                    q = q1;
                    cnt += 1;
                }
            }
            // 处理::xxx
            sum = 0;
            q1 = NULL;
            q = p + 2;
            l1 = strlen(q);
            l1tmp = 0;
            for (i = 0; i < l1; i++) {
                if (*(p + 1 + l1 -i) != ':') {
                    q1 = strrchr(q, ':');
                    *(ipcom + len - i -1 - sum) = *(p + 1 + l1 -i);
                } else {
                    l1tmp = strlen(q);
                    sum += (4 - (l1tmp - (q1 - q) -1));
                    *q1 = '\0';
                    cnt += 1;
                }
            }
        } else {
            
        }
    } else {
        // 不包含::
        l2 = strlen(ipv6);
        sum = 0;
        q1 = NULL;
        q = ipv6;
        for (i = 0; i < l2; i++) {
            if (*(ipv6 + i) != ':') {
                q1 = strchr(q, ':');
                if (q1 == NULL) {
                    q1 = strchr(q, '\0');
                }
                *(ipcom + (4 - (q1 - q)) + sum  + i) = *(ipv6 + i);
            } else {
                sum += (4 - (q1 - q));
                q1 = q1 + 1;
                q = q1;
                cnt += 1;
            }
        }
    } 
    
    memcpy(complete_v6, ipcom, 40);
    return 0;
}

static int compareIpv6(char *ipv6_1, char *ipv6_2)
{
    struct in6_addr ip1;
    struct in6_addr ip2;
    int result = 0;

    inet_pton(AF_INET6, ipv6_1, &ip1);
    inet_pton(AF_INET6, ipv6_2, &ip2);

    result = memcmp(&ip1, &ip2, sizeof(struct in6_addr));
    return result;
}

static int isVaildChar(char c)
{
	if(c >= 'A' && c <= 'F')
		return 0;
	if(c >= 'a' && c <= 'f')
		return 0;
	if(c >= '0' && c <= '9')
		return 0;
	
	return -1;
}

static int IsProtocalIPV6(char *IP){
    char *tmpIp = IP;           
    int sectorCount = 0;    //小段的字符数量，注意这里和v4不一样
    int sector = 0;         //用于存储输入的IP有多少段
	
    complete_ipv6(IP,IP);
    do{
        if(*tmpIp == ':' || *tmpIp == '\0') {
            if(sectorCount == 0 || sectorCount > 4) return -1;  //小段的字符数量为0或者大于4都是非法
            sectorCount = 0;                //清空小段字符数量，为下一次做准备
            sector++;                       //段+1
            if(*tmpIp == '\0') break;       //如果已经到达末尾，则跳出循环
            tmpIp++;                        //跳过:
        }
        if(isVaildChar(*tmpIp) == 0) {       //判断字符是否为合法:0-9 a-f A-F
            sectorCount++;
        }else{                              //异常字符则退出函数
            return -1;
        }
    }while(*tmpIp++ != '\0');
 
    if(sector != 8)     //判断是否有8段，多或少都为非法
        return -1;
    return 0;
}

static uint32_t convert_endian(uint32_t x)
{
	uint32_t result = ((x & 0xff000000) >> 24) | ((x & 0x00ff0000) >> 8) |
	((x & 0x0000ff00) << 8) | ((x & 0x000000ff) << 24);
	return result;
}

static void ipv6_to_integers(const char *ipv6_str, uint32_t *part1, uint32_t *part2, uint32_t *part3,
		      uint32_t *part4) {
	struct in6_addr addr;
	inet_pton(AF_INET6, ipv6_str, &addr);
 
	uint32_t *numbers = (uint32_t *)&addr.s6_addr;
	*part1 = convert_endian(numbers[0]);
	*part2 = convert_endian(numbers[1]);
	*part3 = convert_endian(numbers[2]);
	*part4 = convert_endian(numbers[3]);
} 

static uint32_t ip_to_u32(char *str)
{
	int i, slen = 0;
	int numTemp = 0;
	int dotNum = 0;
	int bitNum = 0;
	uint32_t ipNum = 0;
	if (str == NULL) {
		printf("str == NULL\n");
		return -1;
	}
	slen = strlen(str);
	if (slen < 7 || slen > 15) {
		// printf("strlen(str) < 7 or strlen(str)>15\n");
		return -1;
	}
	if (str[0] == '.' || str[slen - 1] == '.') {
		printf("ip format error\n");
		return -1;
	}
	for (i = 0; i < slen; i++) {
		if (str[i] == '.') {
			if (str[i - 1] == '.') {
				printf("pStr == . format error\n");
				return -1;
			}
			if (++dotNum > 3) { /* judge the number of "." */
				printf("dot number over 3\n");
				return -1;
			}
			if (numTemp > 255) {
				printf("numTemp > 255, numTemp :%d\n", numTemp);
				return -1;
			}
			ipNum |= numTemp;
			numTemp = 0;
			ipNum <<= 8;
			bitNum = 0;
			continue;
		}
		if (str[i] < '0' || str[i] > '9') {
			printf("ip addr error, str :%c\n", str[i]);
			return -1;
		}
		numTemp = numTemp * 10 + (str[i] - '0');
		if (++bitNum > 3) {
			printf("ip addr bit number error\n");
			return -1;
		}
	}
	if (dotNum != 3) {
		printf("dotNum != 3\n");
		return -1;
	}
	ipNum |= numTemp;
	// printf("ipNUm :%lx\n", ipNum);
	return ipNum;
}

static void convert_protocol(char *protocol, uint32_t &protocl_dex)
{
	protocl_dex = 0;
	if (strstr(protocol, "icmpv6") != NULL) {
		protocl_dex = 0x0000003a;
	} else if (strstr(protocol, "icmp") != NULL) {
		protocl_dex = 0x00000001;
	} else if (strstr(protocol, "tcp") != NULL) {
		protocl_dex = 0x00000006;
	} else if (strstr(protocol, "udp") != NULL) {
		protocl_dex = 0x00000011;
	} else if (strstr(protocol, "ah") != NULL) {
		protocl_dex = 0x00000033;
	} else if (strstr(protocol, "esp") != NULL) {
		protocl_dex = 0x00000032;
	}
}

static int brust_data(list<message_t> &all_data_info, char *p, int dex, nlmsghdr *nlh)
{
	list<message_t> data_info;
	int count = 0;
	int index = 0;

	for (auto it = all_data_info.begin(); it != all_data_info.end(); it++) {
		if (index < MAX_SEND * dex) {
			index++;
			continue;
		}
		else if (index == MAX_SEND * (dex + 1)) {
			break;
		}

		message_t sigle_data = *it;
		memcpy(p + count * (HEADER_SIZE + MESSEAGE_LENGTH), (char *)nlh, HEADER_SIZE);
		memcpy(p + count * (HEADER_SIZE + MESSEAGE_LENGTH) + HEADER_SIZE, &sigle_data, MESSEAGE_LENGTH);
		index++;
		count++;
	}
	return count;
}

static list<message_t> get_ops(char *send_data, configure_t *cfg)
{
	FILE *fp;
	fp = fopen(cfg->filename, "r");
	char ops[10][MAX_PAYLOAD];
	char row_data[MAX_PAYLOAD];
	char cut_string[MAX_PAYLOAD];
	uint32_t protocol_no;
	uint32_t src_ip = 0;
	uint32_t dest_ip = 0;
	int index = 0;
	uint32_t conver_ip[2];
	int dex = 0;
	int i = 0;
	int j = 0;
	char *p;
	list<message_t> data_info;

	if (fp == NULL) {
		printf("file %s read failed", cfg->filename);
		return data_info;
	}

	while (fgets(row_data, 1024, fp) != NULL) {
		message_t single_data;
		dex = 0;
		strcpy(cut_string, row_data);
		p = strtok(cut_string, PARA_KEY_STRING);
		strcpy(ops[dex], p);
		while (p != NULL) {
			dex++;
			p = strtok(NULL, PARA_KEY_STRING);
			if (p != NULL)
				strcpy(ops[dex], p);
			if (dex > 3) {
				printf("please check parameter number per row");
				data_info.clear();
				break;
			}
		}
		conver_ip[0] = ip_to_u32(ops[0]);
		conver_ip[1] = ip_to_u32(ops[1]);
		if(conver_ip[0] != -1)
		{
		if(conver_ip[0] > conver_ip[1]) {
			src_ip = conver_ip[0];
			dest_ip = conver_ip[1];
		} else {
			src_ip = conver_ip[0];
			dest_ip = conver_ip[1];
		}
		if (src_ip != -1) {
			single_data.src_ip[3] = src_ip;
			single_data.src_ip[0] = 0;
			single_data.src_ip[1] = 0;
			single_data.src_ip[2] = 0;
		} else {
			continue;
		}

		if (dest_ip != -1) {
			single_data.dest_ip[3] = dest_ip;
			single_data.dest_ip[0] = 0;
			single_data.dest_ip[1] = 0;
			single_data.dest_ip[2] = 0;
		} else {
			continue;
		}

		convert_protocol(ops[2], protocol_no);
		if (protocol_no != 0) {
			single_data.prot_type = protocol_no;
		} else {
			printf("please check protocol!");
			data_info.clear();
			return data_info;
		}

		if (cfg->infName != NULL)
			memcpy(single_data.port, cfg->infName, 16);

		data_info.push_back(single_data);
		index++;
		}
		else if(IsProtocalIPV6(ops[0]) != -1 && IsProtocalIPV6(ops[1]) != -1)
		{
			if(compareIpv6(ops[0], ops[1]) < 0)
			{
				ipv6_to_integers(ops[0], &single_data.src_ip[0], &single_data.src_ip[1], &single_data.src_ip[2], &single_data.src_ip[3]);
				ipv6_to_integers(ops[1], &single_data.dest_ip[0], &single_data.dest_ip[1], &single_data.dest_ip[2], &single_data.dest_ip[3]);
			}
			else {
				ipv6_to_integers(ops[1], &single_data.src_ip[0], &single_data.src_ip[1], &single_data.src_ip[2], &single_data.src_ip[3]);
				ipv6_to_integers(ops[0], &single_data.dest_ip[0], &single_data.dest_ip[1], &single_data.dest_ip[2], &single_data.dest_ip[3]);
			}
			if (cfg->infName != NULL)
				memcpy(single_data.port, cfg->infName, 16);

			convert_protocol(ops[2], protocol_no);
			if (protocol_no != 0) {
				single_data.prot_type = protocol_no;
			} else {
				printf("please check protocol!");
				data_info.clear();
				return data_info;
			}	
			data_info.push_back(single_data);

		}
		
		index++;

	}

	fclose(fp);
	return data_info;
}

static void usage(char *appName)
{
	printf("\r\n"
	       "Usage: %s [-h] -i <interface> -f <package> \r\n"
	       " -h             Print this message and exit\r\n"
	       " -a             add whitelist\r\n"
	       " -r             rm whitelist\r\n"
	       " -i [device]    Interface to use, for example 'eth0'\r\n"
	       " -f [filename]  Specify the Settings filename, for example 'test.xml'\r\n"
	       " If none of l/a/n/i specified, then all parts will be updated.\r\n",
	       appName);
}

static int32_t configure_parse_ops(configure_t *cfg, uint32_t argc, char *argv[])
{
	uint32_t c;
	char *appName = argv[0];

	if (argc < 2) {
		usage(appName);
		return -1;
	}

	cfg->hasInf = false;
	cfg->hasFile = false;
	cfg->addData = false;
	cfg->rmData = false;

	while (EOF != (c = getopt(argc, argv, "larnphi:f:"))) {
		switch (c) {
		case 'i':
			sprintf(cfg->infName, "%s", optarg);
			cfg->hasInf = true;
			break;
		case 'f':
			sprintf(cfg->filename, "%s", optarg);
			cfg->hasFile = true;
			break;
		case 'a':
			cfg->addData = true;
			break;
		case 'r':
			cfg->rmData = true;
			break;		
		case 'h':
			usage(appName);
			cfg->appName = appName;
			return 0;
		default:
			printf("Unsupported argument!\n");
			return -1;
		}
	}

	if (!(cfg->hasInf && cfg->hasFile)) {
		printf("Please provide interface through -i, provide file through -f\n");
		return -1;
	}
	if (!(cfg->addData || cfg->rmData)) {
		printf("Please provide add whitelist -a, or provide remove whitlist through -r\n");
		return -1;
	}
	if ((cfg->addData && cfg->rmData)) {
		printf("Please only provide add whitelist -a, or provide remove whitlist through -r\n");
		return -1;
	}

	return 0;
}

int32_t send_whitelist(int argc, char *argv[])
{
	configure_t *cfg = &configure;

	if (0 != configure_parse_ops(cfg, argc, argv)) {
		return -1;
	}

	struct sockaddr_nl src_addr, dest_addr;
	struct nlmsghdr *nlh = NULL;
	int sock_fd, retval;
	struct msghdr msg;
	struct iovec iov;
	int state;
	char send_data[MAX_PAYLOAD];
	int state_smg = 0;
	int dex = 0;
	int count = 1;
	struct netlink_state send_result;
	int wait_recvmsg = 1;
	list<message_t> data_info = get_ops(send_data, cfg);
	if (data_info.size() == 0) {
		printf("please check param!");
		return 0;
	}
	sock_fd = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_TEST);
	if (sock_fd == -1) {
		printf("error getting socket: %s\n", strerror(errno));
		return -1;
	}
	/* To prepare binding */
	memset(&src_addr, 0, sizeof(src_addr));
	src_addr.nl_family = AF_NETLINK;
	src_addr.nl_pid = getpid();
	src_addr.nl_groups = 0;
	/* Bind */
	retval = bind(sock_fd, (struct sockaddr *)&src_addr, sizeof(src_addr));
	if (retval < 0) {
		printf("bind failed: %s", strerror(errno));
		close(sock_fd);
		return -1;
	}
	/* To orepare create mssage */
	nlh = (struct nlmsghdr *)malloc(NLMSG_SPACE(FACT_PAYLOAD));
	if (!nlh) {
		printf("malloc nlmsghdr error!\n");
		close(sock_fd);
		return -1;
	}
	memset(&dest_addr, 0, sizeof(dest_addr));
	dest_addr.nl_family = AF_NETLINK;
	dest_addr.nl_pid = 0;
	dest_addr.nl_groups = 0;

	nlh->nlmsg_len = NLMSG_SPACE(MESSEAGE_LENGTH);
	nlh->nlmsg_pid = getpid();
	nlh->nlmsg_flags = 1;
	if(cfg->addData)
		nlh->nlmsg_type = WhiteList_Add;
	else
		nlh->nlmsg_type = WhiteList_RM;
	/* send message */
	dex = 0;
	while (count) {
		nlh->nlmsg_len = NLMSG_SPACE(MESSEAGE_LENGTH);
		nlh->nlmsg_pid = getpid();
		nlh->nlmsg_flags = 1;
		if(cfg->addData)
			nlh->nlmsg_type = WhiteList_Add;
		else
			nlh->nlmsg_type = WhiteList_RM;

		count = brust_data(data_info, send_data, dex, nlh);
		if (!count) {
			break;
		}
		iov.iov_base = (void *)send_data;
		iov.iov_len = ((HEADER_SIZE + MESSEAGE_LENGTH) * count);
		/* Create message */
		memset(&msg, 0, sizeof(msg));
		msg.msg_name = (void *)&dest_addr;
		msg.msg_namelen = sizeof(dest_addr);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		state_smg = sendmsg(sock_fd, &msg, 0);
		// printf("state_smg:%d\n", state_smg);
		if (state_smg == -1) {
			printf("get error sendmsg = %s\n", strerror(errno));
			close(sock_fd);
			free(nlh);
			return -1;
		}
		memset(nlh, 0, NLMSG_SPACE(FACT_PAYLOAD));
		// receive message
		printf("waiting received!\n");
		while (wait_recvmsg) {
			state = recvmsg(sock_fd, &msg, 0);
			if (state <= 0) {
				printf("waiting recvmsg! < 1\n");
			}

			if (state > 0) {
				send_result = *(struct netlink_state*) NLMSG_DATA((nlmsghdr *)send_data);
				for (int i = 0; i < 64; i++) {
					if ((send_result.writeData >> i) & 0x1) {
						printf("the row number:%d  configuration data erro "
						       "\n",
						       i + 1 + dex * MAX_SEND);
					}
				}
				wait_recvmsg = 0;
			}
		}
		dex++;
	}
	printf("send whitelist success!\n");
	close(sock_fd);
	free(nlh);
	return 0;
}
