#include "crc32.h"

static uint32_t crcTable[CRC32_TABLE_SIZE];  // 1KB

void Crc32Init(void)
{
    uint32_t remainder;
    uint32_t dividend;
    int32_t  bit;

    for (dividend = 0U; dividend < CRC32_TABLE_SIZE; ++dividend) {
        remainder = dividend;

        for (bit = 8; bit > 0; --bit) {
            if (0 != (remainder & 1U)) {
                remainder = (remainder >> 1) ^ CRC32_REVERSED_POLYNOMIAL;
            }
            else {
                remainder >>= 1;
            }
        }

        crcTable[dividend] = remainder;
    }
}

uint32_t Crc32(const void *pMessage, uint32_t bytes)
{
    uint32_t remainder = CRC32_INITIAL_REMAINDER;
    const uint8_t *pBuffer = (const uint8_t *)pMessage;
    uint8_t  idx;

    while (0 < bytes--) {
        idx = (uint8_t)(*pBuffer++ ^ remainder);
        remainder = crcTable[idx] ^ (remainder >> 8);
    }

    return remainder ^ CRC32_FINALIZE_REMAINDER;
}

uint32_t Crc32_Part(uint32_t crc32, const void *pMessage, uint32_t bytes)
{
    uint32_t remainder = crc32 ^ CRC32_INITIAL_REMAINDER;
    const uint8_t *pBuffer = (const uint8_t *)pMessage;
    uint8_t  idx;

    while (0 < bytes--) {
        idx = (uint8_t)(*pBuffer++ ^ remainder);
        remainder = crcTable[idx] ^ (remainder >> 8);
    }

    return remainder ^ CRC32_FINALIZE_REMAINDER;
}

